__all__ = ()

COMPONENT_SUB_COMPONENT_LIMIT = 5
COMPONENT_LABEL_LENGTH_MAX = 80
COMPONENT_CUSTOM_ID_LENGTH_MAX = 100
COMPONENT_OPTION_LENGTH_MIN = 1
COMPONENT_OPTION_LENGTH_MAX = 25
COMPONENT_OPTION_MIN_VALUES_MIN = 0
COMPONENT_OPTION_MIN_VALUES_MAX = 15
COMPONENT_OPTION_MAX_VALUES_MIN = 1
COMPONENT_OPTION_MAX_VALUES_MAX = 25
COMPONENT_MAX_LENGTH_MIN = 0
COMPONENT_MAX_LENGTH_MAX = 4000
COMPONENT_MIN_LENGTH_MIN = 0
COMPONENT_MIN_LENGTH_MAX = 4000
COMPONENT_TITLE_LENGTH_MIN = 1
COMPONENT_TITLE_LENGTH_MAX = 45
